@@ -13,7 +13,7 @@ from .types import FileCategory, File, ReadResult
 
 from .__util_storage import Storage
 from .__util_datalake import Datalake
-from .__util_es_datalake import es_datalake_search_eql
+from .__util_es_datalake import es_datalake_search_eql, es_hit_to_file_pointer
 from .__util_log import Log
 from .__util_merge import merge_arrays, merge_objects
 from .__util_metadata import FIELDS
@@ -126,6 +126,19 @@ class Context:
         ...     df = pd.read_sql_query('SELECT * FROM foo', con)
         """
 
+        if len(file.keys()) == 1 and file['fileId'] is not None:
+            es_files = self.search_eql({
+                'size' : 1,
+                'query': {
+                    'term': {
+                        'fileId': {'value': file['fileId']}
+                    }
+                }
+            }, returns='filePointers')
+            if len(es_files) == 0:
+                raise Exception(f"File with fileId {file['fileId']} not found!")
+            file = es_files[0]
+
         return self._datalake.read_file(file, form)
 
     @wrap_log('context.write_file')
@@ -360,9 +373,14 @@ class Context:
         
         return new_file
 
+    
+
     @wrap_log('context.search_eql')
-    def search_eql(self, payload):
-        return es_datalake_search_eql(payload)
+    def search_eql(self, payload, returns='raw'):
+        search_res = es_datalake_search_eql(payload)
+        if returns == 'filePointers':
+            return list(map(es_hit_to_file_pointer, search_res['hits']['hits']))
+        return search_res
 
 
 def output_response(storage, response, correlation_id):
@@ -480,6 +498,10 @@ def run(input, context_from_arg, func, correlation_id, func_dir,
     except Exception as e:
         log.log(e)
 
+        if 'SIGKILL' in str(e):
+            log.log({ 'level': 'error', 'message': 'Got SIGKILL - possible OOM'})
+            os._exit(137)
+
         if store_output:
             output_response(storage, { 'type': FAILED, 'error': str(e) }, correlation_id)
 
