@@ -21,6 +21,7 @@ DISABLE_GZIP = os.environ.get('DISABLE_GZIP')
 ENV = os.environ.get('ENV')
 AWS_REGION = os.environ.get('AWS_REGION')
 MULTIPART_SIZE = 100 * 1024 * 1024 # don't use multipart if file size is smaller than 100MB
+LARGE_FILE_SIZE_THRESHOLD_BYTES = 5 * 2**30 # We have to handle files that are > 5 GB differently
 
 def get_kms_key_name(org_slug):
     return f'alias/customer-key-{ENV}-{org_slug}'
@@ -265,26 +266,57 @@ class Datalake:
                 sse_kms_key_id=get_kms_key_name(org_slug)
             )
 
+        version_id = ''
         if hasattr(content, 'read'):
             response = S3FileobjUploader(
                 self.s3, content, params, { 'disable_gzip': DISABLE_GZIP }).upload()
+            # fakeS3 does not return VersionId, so use '' to avoid an exception
+            version_id = response.get('VersionId', '')
+            print({'level': 'debug', 'message': 'file created', 'response': response})
         else:
             if not DISABLE_GZIP:
                 if (isinstance(content, str)):
                     content = content.encode()
                 content = gzip.compress(content)
-            response = self.s3.put_object(Body=content, **params)
-        print({ 'level': 'debug', 'message': 'file created', 'response': response })
+
+            file_size = len(content) or 0
+            if file_size >= LARGE_FILE_SIZE_THRESHOLD_BYTES:
+                print({'level': 'debug', 'message': 'writing a 5+ GB file to s3'})
+
+                # the upload_fileobj method requires a file like object that implements read. This means we have to
+                # wrap content into an in memory stream. We will use BytesIO instead of StringIO since the metadata has
+                # to be encoded. This means that if content is a string, we have to convert it to bytes first.
+                file_like_obj = None
+                if type(content) is str:
+                    file_like_obj = io.BytesIO(content.encode('utf-8'))
+                else:
+                    # content was already bytes so we should not try to encode it again
+                    file_like_obj = io.BytesIO(content)
+                try:
+                    extra_args = self.extract_extra_args_from_params(params)
+
+                    self.s3.upload_fileobj(Fileobj=file_like_obj, Bucket=bucket, Key=file_key, ExtraArgs=extra_args)
+                    version_id = self.get_latest_obj_version(bucket, file_key)
+                    print({'level': 'debug', 'message': 'file created'})
+
+                except Exception as e:
+                    raise Exception("encountered an error updating the metadata because of {}".format(e))
+                finally:
+                    file_like_obj.close()
+            else:
+                response = self.s3.put_object(Body=content, **params)
+                # fakeS3 does not return VersionId, so use '' to avoid an exception
+                version_id = response.get('VersionId', '')
+                print({ 'level': 'debug', 'message': 'file created', 'response': response })
+
         result_file = {
             'type': 's3file',
             'bucket': bucket,
             'fileKey': file_key,
             'fileId': file_id,
-            # fakeS3 does not return VersionId, so use '' to avoid an exception
-            'version': response.get('VersionId', '')
+            'version': version_id
         }
 
-
         return result_file
 
     def update_metadata_tags(self, context, file, custom_meta, custom_tags, options = {}):
@@ -298,7 +330,8 @@ class Datalake:
             raise Exception('no FILE_ID in meta!')
 
         if (not custom_meta) and (not custom_tags):
-            raise Exception('No metadata or tags provided')
+            print({'level': 'debug', 'message': 'no action taken because no metadata or tags provided'})
+            return file
 
         isASCII = lambda s: s and isinstance(s, str) and bool(re.match(r'^[\x00-\x7F]*$', s))
 
@@ -365,15 +398,38 @@ class Datalake:
         }
 
         params = {k:v for k,v in params.items() if v is not None}
-        response = self.s3.copy_object(**params)
+        # fakeS3 does not return VersionId, so use '' to avoid an exception when we try to retrieve from the head
+        # object or directly from the response object
+        updated_version = ''
+
+        # boto3.client.copy_object() has a maximum size of 5 GB. If the file exceeds this limit, we have to use the
+        # boto3.copy() method that will do the multipart upload for us.
+        file_size = head.get('ContentLength') or 0
+
+        if file_size >= LARGE_FILE_SIZE_THRESHOLD_BYTES:
+            print({'level': 'debug', 'message': 'updating the metadata of a 5+ GB file'})
+            extra_args = self.extract_extra_args_from_params(params)
+            copy_source = {'Bucket': bucket, 'Key': file_key}
+
+            # unlike copy_object(), the copy() function does not return a response object. It will raise an error
+            # instead. Additionally, we will have to make an additional request to get the versionId if it exists
+            try:
+                self.s3.copy(CopySource=copy_source, Bucket=bucket, Key=file_key, ExtraArgs=extra_args)
+                updated_version = self.get_latest_obj_version(bucket, file_key)
+            except Exception as e:
+                raise Exception("encountered an error updating the metadata because of {}".format(e))
+        else:
+            # For most cases when files are less than 5 GB in size. The versionId comes from the response object of
+            # the copy_object() function
+            response = self.s3.copy_object(**params)
+            updated_version = response.get('VersionId', '')
 
         return {
             'type': 's3file',
             'bucket': bucket,
             'fileKey': file_key,
             'fileId': file_id,
-            # fakeS3 does not return VersionId, so use '' to avoid an exception
-            'version': response.get('VersionId', '')
+            'version': updated_version
         }
 
     def write_ids(self, context, content_obj, file_suffix, raw_file, file_meta, ids, source_type, file_category, labels):
@@ -434,3 +490,23 @@ class Datalake:
         response = self.s3.put_object(Body=json.dumps(labels), **params)
         print({ 'level': 'debug', 'message': 'labels file created', 'response': response })
 
+    # extra_args can only have specific key names. In this case, it cannot contain the
+    # keys: [Bucket, Key, CopySource]. These keys are already included in the copy() function signature.
+    def extract_extra_args_from_params(self, params):
+        extra_args = {}
+        for k, v in params.items():
+            if k in ['Bucket', 'Key', 'CopySource']:
+                continue
+            extra_args[k] = v
+        return extra_args
+
+    # We can't use s3.get_s3_head(file) because the file may have the old version before the copy command was called
+    # So we need to make a direct call to head_object without the versionId to get the latest version of the file
+    def get_latest_obj_version(self, bucket, file_key):
+        try:
+            new_head_obj = self.s3.head_object(Bucket=bucket, Key=file_key)
+            return new_head_obj.get('VersionId', '')
+        except Exception as e:
+            print({'level': 'error', 'message': 'could not retrieve the latest version of the file because {}'.format(e)})
+            return ''
+
